/*
** Module   :DISPATCH.CPP
** Abstract :Keyboard events dispatcher
**
** Copyright (C) Sergey I. Yevtushenko
**
** Log: Thu  26/03/1998     Created
*/

#include <string.h>

#include <boxcoll.h>
#include <keynames.h>
#include <version.h>

#define INCL_DOS
#include <os2.h>

//-----------------------------------------
// Dispatcher
//-----------------------------------------

void EditBoxCollection::Dispatcher(KeyInfo& k, int iMode)
{
    lock();
    track_beg();

    if((k.skey & 0x00FF) == kbEnter &&
    	!(k.skey & (shAlt | shCtrl | shShift)))
            k.skey &= ~shIsCtrl;

    char* pre_key  = (iMode) ? keys.GetDef("kbBeforeKey"):0;
    char* post_key = (iMode) ? keys.GetDef("kbAfterKey"):0;

    if(pre_key)
    {
        track_recording(pre_key, 0);
        play_macro(pre_key);
    }

    if(k.skey & shIsCtrl)
    {
        char* prog = keys.GetDef(k.KeyName);

        if(prog)
        {
            track_recording(prog, 0);
            play_macro(prog);
        }
    }
    else
    {
        track_recording(0, (char)k.key);
        usual_key(k);
        doFlashBracket(*this, 0);
    }

    if(post_key)
    {
        track_recording(post_key, 0);
        play_macro(post_key);
    }

    track_end();
    unlock();
}

void EditBoxCollection::play_macro(char *macro)
{
    if(!macro)
        return;

    KeyInfo k;

    while(*macro)
    {
        if(*macro == FUNC_ESCAPE)
        {
            macro++;

            if(PEditProc(macro))
            {
            	PEditProc proc;
            	memcpy(&proc, macro, sizeof(PEditProc));
            	macro += sizeof(PEditProc);
            	macro += proc(*this, macro);
            }
            else
            	return;
        }
        else
        {
            k.key = *macro++;
            usual_key(k);
            doFlashBracket(*this, 0);
        }
    }
}

void EditBoxCollection::track_recording(char *macro, char chr)
{
    if(!recording)
        return;

    macro_rec *tmp = new macro_rec(macro, chr, 0);

    if(!head)
        head = tmp;

    if(tail)
        tail->next = tmp;
    tail = tmp;
}

void EditBoxCollection::clear_recording()
{
    macro_rec *tmp;

    while(head)
    {
        tmp = head->next;
        delete head;
        head = tmp;
    }
    head = tail = 0;
    recording = 0;
}

char* EditBoxCollection::track_recording_done()
{
    macro_rec *tmp;
    int len   = 0;
    char *out = 0;

    for(tmp = head; tmp; tmp = tmp->next)
    {
        if(tmp->macro)
            len += strlen(tmp->macro);
        else
            len++;
    }

    out = new char[len + 1];

    char *ptr = out;

    for(tmp = head; tmp; tmp = tmp->next)
    {
        if(tmp->macro)
        {
            char *str = tmp->macro;

            while(*str)
            {
                if(*str == FUNC_ESCAPE)
                {
                    union
                    {
                        PEditProc proc;
                        char data[sizeof(PEditProc)];
                    }
                    cvt;

                    memcpy(cvt.data, str + 1, sizeof(cvt.data));

                    if(cvt.proc == doMacroRecEnd  ||
                       cvt.proc == doMacroRecStart)
                    {
                        str += sizeof(cvt.data) + 1;
                    }
                    else
                    {
                        *ptr++ = *str++;
                        memcpy(ptr, str, sizeof(cvt.data));
                        ptr += sizeof(cvt.data);
                        str += sizeof(cvt.data);
                    }
                }
                else
                    *ptr++ = *str++;
            }
        }
        else
            *ptr++ = tmp->chr;
    }
    *ptr = 0;

    return out;
}

void EditBoxCollection::start_recording()
{
    recording = 1;
}

