/*
** Module   :RX.CPP
** Abstract :
**
** Copyright (C) Sergey I. Yevtushenko
**
** Log: Fri  10/04/1998 Created
**
*/

#include <string.h>

#include <boxcoll.h>
#include <stddlg.h>
#include <keynames.h>
#include <fio.h>
#include <_search.h>
#include <version.h>
//#include <_regex.h>

#define INCL_REXXSAA
#define INCL_DOS

#ifdef __EMX__
extern "C" {
#include <os2emx.h>
}
#else
#include <rexxsaa.h>
#endif

//-----------------------------------------
// Data Types
//-----------------------------------------

typedef struct
{
    char *pszName;
    PFN pfnFunc;
} FuncListItem;

//-----------------------------------------
// Function Prototypes
//-----------------------------------------

ULONG APIENTRY fedNoparm  (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedGet     (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedSet     (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedFind    (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedInsChar (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedGetChar (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedOpen    (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedJumpList(PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedMsgBox  (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);

ULONG APIENTRY fedVarSet  (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedVarGet  (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedVarDrop (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);

ULONG APIENTRY fedInput   (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedGetClip (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedSetClip (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);

ULONG APIENTRY fedBind    (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedMenu    (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);
ULONG APIENTRY fedSendKey (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);

ULONG APIENTRY fedRxMatch (PCSZ, ULONG, PRXSTRING, PCSZ, PRXSTRING);

int makeRC   (PRXSTRING, LONG );
int makeRCstr(PRXSTRING, char*);
int makeRCstr(PRXSTRING, int  );

APIRET GetRexxVariable(char* name, char* value, LONG len);
APIRET SetRexxVariable(char* name, char* value, LONG len);

static char* mk_var(char *buff, char *beg, int num, char *end, int mode);

//-----------------------------------------
// Static and Extern Data
//-----------------------------------------

extern EditBoxCollection Editor;
extern JumpList JList[];
extern DispatchEntry dispatch_list[];

FuncListItem FunctionList[]=
{
    { "fedBind"        , (PFN)fedBind    },
    { "fedFillJumpList", (PFN)fedJumpList},
    { "fedFind"        , (PFN)fedFind    },
    { "fedGet"         , (PFN)fedGet     },
    { "fedGetChar"     , (PFN)fedGetChar },
    { "fedGetClip"     , (PFN)fedGetClip },
    { "fedInput"       , (PFN)fedInput   },
    { "fedInsChar"     , (PFN)fedInsChar },
    { "fedMenu"        , (PFN)fedMenu    },
    { "fedMsgBox"      , (PFN)fedMsgBox  },
    { "fedOpenFile"    , (PFN)fedOpen    },
    { "fedOpenJumpList", (PFN)fedJumpList},
    { "fedSendKey"     , (PFN)fedSendKey },
    { "fedSet"         , (PFN)fedSet     },
    { "fedSetClip"     , (PFN)fedSetClip },
    { "fedVarDrop"     , (PFN)fedVarDrop },
    { "fedVarGet"      , (PFN)fedVarGet  },
    { "fedVarSet"      , (PFN)fedVarSet  },
    { "fedRxMatch"     , (PFN)fedRxMatch },
};

//-----------------------------------------
// API's
//-----------------------------------------

void InitREXX(void)
{
	int i;

	//Regular REXX API's
	for(i = 0; i < (sizeof(FunctionList)/sizeof(FunctionList[0])); i++)
	{
		RexxRegisterFunctionExe((PCSZ)FunctionList[i].pszName,
#ifdef __EMX__
								(RexxFunctionHandler*)FunctionList[i].pfnFunc);
#else
								FunctionList[i].pfnFunc);
#endif
	}

	char cName[64] = "fed";

	for(i = 0; dispatch_list[i].key; i++)
	{
		if(dispatch_list[i].key[0] == '-')
			continue;

		strcpy(cName + 3, dispatch_list[i].key);

		RexxRegisterFunctionExe((PCSZ)cName,
#ifndef __EMX__
		(PFN)
#endif
		fedNoparm);
	}
}

//-----------------------------------------------------------------------
// REXX API
//-----------------------------------------------------------------------

//-----------------------------------------
// fedRxMatch(string, regexp[, [flags] [, pat0[, pat1[...]]]])
//
// string - source string
// regexp - regular expression
// flags  - search options ('I' - ignore case)
// pat0   - variable which will contain full matching string,
// pat1...patN - variables where matching subexpressions will be stored
//-----------------------------------------

ULONG APIENTRY fedRxMatch(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    int rc = 0;
    int ignore_case = 0;

    //At least two arguments must be present
    if(lCnt < 2)
        return 40;

    if(lCnt > 2 && argv[2].strptr)
    {
    	char* s = (char*)argv[2].strptr;

    	while(*s)
    		if(__to_upper(*s++) == 'I')
    			ignore_case = 1;
    }

    RXSearch flt;

    do
    {
        rc = flt.init((char*)argv[1].strptr, ignore_case);

        if(rc)
        {
            rc = -1;
            break;
        }

        int len = 0;
        char* start = flt.search((char*)argv[0].strptr, len);

        if(!start)
        {
        	rc = 0;
        	break;
        }

        rc = 1; //Match found

        for(int i = 3; i < lCnt; i++)
        {
			int beg = 0;
			int end = 0;

			flt.get_match(i - 3, beg, end);

			SetRexxVariable((char*)argv[i].strptr, (char*)&argv[0].strptr[beg], end - beg);
        }
    }
    while(0);

    return makeRC(RC, rc);
}

ULONG APIENTRY fedInput(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    char *hdr = 0;
    char *res = 0;
    int r = 5;
    int c = 5;
    int rc;

    if(lCnt > 3)
        return 40;

    if(lCnt > 0)
        hdr = (char*)argv[0].strptr;

    if(lCnt > 1)
        r = s2u((char*)argv[1].strptr);

    if(lCnt > 2)
        c = s2u((char*)argv[2].strptr);

    rc = AskString(r, c, &res, hdr);

    if(!rc)
    {
        int Len = strlen(res);

        makeRCstr(RC, Len + 1);

        strcpy((char*)RC->strptr, res);
        RC->strlength = Len;

        delete res;
        return 0;
    }

    RC->strlength = 0;
    return 0;
}

ULONG APIENTRY fedGetClip(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt)
        return 40;

    if(!Clipboard)
    {
        RC->strlength = 0;
        return 0;
    }

    Clipboard->from_pm();

    char *res = Clipboard->as_text();

    RC->strlength = strlen(res);
    RC->strptr    = (PCH)res;

    return 0;
}

ULONG APIENTRY fedSetClip(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt != 1 || !argv[0].strptr)
        return 40;

    if(!Clipboard)
        return 0;

    Clipboard->from_text((char*)argv[0].strptr);
    Clipboard->to_pm();

    return 0;
}

ULONG APIENTRY fedSendKey(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt != 1 || !argv[0].strlength || !argv[0].strptr)
        return 40;

    Editor.SendKey((char*)argv[0].strptr);

    RC->strlength = 0;
    return 0;
}

ULONG APIENTRY fedBind(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt)
    {
        if(!argv[0].strptr || !argv[0].strlength)
            return 40;

        char *file = _ld_file((char*)argv[0].strptr);

        if(file)
            Editor.load_profile(file);

        _fr_file(file);
    }
    else
        EditBoxCollection::doLoadProfile(Editor, 0);

    RC->strlength = 0;
    return 0;
}

ULONG APIENTRY fedVarSet(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt < 1 || lCnt > 2)
        return 40;

    if(lCnt == 1)
        Editor.get_bindery().delVar((char*)argv[0].strptr);
    else
        Editor.get_bindery().setVar((char*)argv[0].strptr, (char*)argv[1].strptr);

    RC->strlength = 0;
    return 0;
}

ULONG APIENTRY fedVarGet(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt < 1 || lCnt > 2)
        return 40;

    char *var = Editor.get_bindery().getVar((char*)argv[0].strptr);

    if(var)
    {
        int mLen = strlen(var) + 1;

        if(mLen > RC->strlength)
        {
            if(DosAllocMem((PPVOID)&RC->strptr, mLen, PAG_COMMIT | PAG_WRITE))
            {
                RC->strlength = 0;
                return 0;
            }
        }
        strcpy((char*)RC->strptr, var);
    }
    else
    {
        if(lCnt == 1)
        {
            //strcpy((char*)RC->strptr, strupr((char*)argv[0].strptr));
            RC->strptr[0] = 0;
        }
    	else
    	{
        	if(lCnt == 2)
                strcpy((char*)RC->strptr, (char*)argv[1].strptr);
        }
    }

    RC->strlength = strlen((char*)RC->strptr);
    return 0;
}

ULONG APIENTRY fedVarDrop(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt != 1)
        return 40;

    Editor.get_bindery().dropVar((char*)argv[0].strptr);

    RC->strlength = 0;
    return 0;
}

ULONG APIENTRY fedOpen(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt > 1)
        return makeRC(RC, -1);

    if(!lCnt || !argv[0].strptr || !argv[0].strlength ||
       strchr((char*)argv[0].strptr, '?') ||
       strchr((char*)argv[0].strptr, '*'))
    {
        Editor.open();
        return makeRC(RC, 1);
    }
    else
        Editor.doOpenFile((char*)argv[0].strptr);

    return makeRC(RC, 0);
}

ULONG APIENTRY fedMsgBox(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt < 1 || lCnt > 2 || !argv[0].strptr || !argv[0].strlength)
        return makeRC(RC, -1);

    int timeout = 0;

    if(lCnt > 1)
        timeout = s2u((char*)argv[1].strptr);

    MessageBox((char*)argv[0].strptr, 0, timeout);

    return makeRC(RC, 0);
}

ULONG APIENTRY fedNoparm(PCSZ pName, ULONG lCnt, PRXSTRING, PCSZ, PRXSTRING RC)
{
    if(lCnt != 0) //No arguments should be given
        return makeRC(RC, -1);

	PDEntry pFunc = Editor.GetFunction((char*)(pName + 3));

    if(pFunc)
		pFunc->proc(Editor, 0);

    return makeRC(RC, 0);
}

ULONG APIENTRY fedInsChar(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt != 1) //One argument should be given
        return makeRC(RC, -1);

    Editor.play_macro((char*)argv[0].strptr);
    return makeRC(RC, 0);
}

ULONG APIENTRY fedGetChar(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    //If no arguments given,
    //return current string
    //With one argument, return full string
    //With two arguments, return rest of string from given pos
    //With three arguments, return part of string from given pos;

    int Row = Editor.current()->abs_row();
    int Col = 0;
    int Len = -1;

    switch(lCnt)
    {
        case 3:
            if(argv[2].strlength != 0) //Length
            {
                Len = s2u((char*)argv[2].strptr);

                if(Len == 0)
                    return makeRC(RC, -1);
            }

        case 2:

            if(argv[1].strlength != 0) //Col
            {
                Col = s2u((char*)argv[1].strptr);

                if(Col == 0)
                    return makeRC(RC, -2);

                Col--;

                if(Len < 0)
                    Len = -2;
            }

        case 1:

            if(argv[0].strlength != 0) //Row
            {
                Row = s2u((char*)argv[0].strptr);

                if(Row == 0)
                    return makeRC(RC, -3);
            }

            Row--;

        case 0:

            /*
            RC->strlength = (Editor.current()->get_cur_char()) ? 1:0;
            RC->strptr[0] = (char)Editor.current()->get_cur_char();
            RC->strptr[1] = '\0';
            break;
            */

            if(Editor.current()->Count() < Row)
                return makeRC(RC, -1);

            if(Len == -2)
                Len = Editor.current()->line(Row)->len() - Col;

            if(Len == -1)
                Len = Editor.current()->line(Row)->len();

            makeRCstr(RC, Len);

            Editor.current()->line(Row)->get_print(Col, (char*)RC->strptr, Len);
            RC->strlength = Len;
            break;

        default:
            return makeRC(RC, -4);
    }

    return 0;
}


ULONG APIENTRY fedGet(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt != 1 || argv[0].strlength < 1)
        return makeRC(RC, -1);

    //Supported values:
    //  Ins    - Insert mode
    //  Hi     - Hiliting mode
    //  Auto   - Autoindent state
    //   AC      - Autocompletion state
    //  Mark   - Marking state
    //  Row    - Current absolute Row
    //  Col    - Current absolute Col
    //   CH      - Changed status
    //   CR      - Cursor Row
    //   CC      - Cursor Col
    //  CP     - Code Page
    //  Buffer - Current buffer handle
    //  BR     - Marking position start row
    //  BC     - Marking position start col
    //  ? BufferCount - Number of buffers
    //  Name   - Current buffer name
    //  XSize  - Rect size X (Cols)
    //  YSize  - Rect size Y (Rows)
    //  Last   - Last row number
    //  Dir    - Startup dir
    //  WW     - WordWrap Width
    //  WS     - WordWrap Status
    //  WM     - WordWrap Merge
    //  WP     - WordWrap Para

    switch(__to_upper(argv[0].strptr[0]))
    {
        case 'T': return makeRC(RC, iTabWidth);
        case 'I': return makeRC(RC, Editor.current()->get_ins_mode());
        case 'H': return makeRC(RC, Editor.current()->get_hiliting()   );
        case 'A':
            if(argv[0].strlength < 2)
                return makeRC(RC, Editor.current()->get_auto_indent());

            if(__to_upper(argv[0].strptr[1]) == 'C')
                return makeRC(RC, Editor.current()->get_auto_completion());
            break;

        case 'M': return makeRC(RC, Editor.current()->get_mark_state() );
        case 'R': return makeRC(RC, Editor.current()->get_edit_row()   );
        case '?': return makeRC(RC, (ULONG)Editor.Count());
        case 'X': return makeRC(RC, Editor.current()->cols );
        case 'Y': return makeRC(RC, Editor.current()->rows );
        //String return
        case 'N': return makeRCstr(RC, Editor.current()->get_name());
        case 'D': return makeRCstr(RC, StartupDir);

        case 'B':
            if(argv[0].strlength < 2)
                return makeRC(RC, (ULONG)Editor.current());

            switch(__to_upper(argv[0].strptr[1]))
            {
                case 'R':
                    if(!Editor.current()->get_mark_state())
                        return makeRC(RC, Editor.current()->get_edit_row());

                    return makeRC(RC, Editor.current()->get_edit_row_2());

                case 'C':
                    if(!Editor.current()->get_mark_state())
                        return makeRC(RC, Editor.current()->get_edit_col());

                    return makeRC(RC, Editor.current()->get_edit_col_2());
            }
            break;

        //Possible other cases
        case 'C':
            if(argv[0].strlength < 2)
                return makeRC(RC, Editor.current()->get_edit_col());

            switch(__to_upper(argv[0].strptr[1]))
            {
                case 'H': return makeRC(RC, Editor.current()->get_changed());
                case 'R': return makeRC(RC, Editor.current()->get_cur_row());
                case 'C': return makeRC(RC, Editor.current()->get_cur_col());
                case 'P': return makeRCstr(RC, Editor.current()->get_cur_cp());
            }
            break;

        case 'W':
            if(argv[0].strlength != 2)
                break;

            switch(__to_upper(argv[0].strptr[1]))
            {
                case 'W': return makeRC(RC, Editor.current()->ww_get_width());
                case 'S': return makeRC(RC, (Editor.current()->ww_get() & WW_STATE) ? 1:0);
                case 'M': return makeRC(RC, (Editor.current()->ww_get() & WW_MERGE) ? 1:0);
                case 'P': return makeRC(RC, (Editor.current()->ww_get() & WW_LONG ) ? 1:0);
            }
            break;


        case 'L':
            if(__to_upper(argv[0].strptr[1]) == 'C')  //Last char
                return makeRCstr(RC, kiLastKey.KeyName);

            return makeRC(RC, Editor.current()->Count());

        default:
            break;
    }
    return makeRC(RC, -2);
}

ULONG APIENTRY fedSet(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt < 1 || argv[0].strlength < 1)
        return makeRC(RC, -1);

    //Supported values:
    //  Ins    - Insert mode               (flip)
    //  Hi     - Hiliting mode             (flip)
    //  Auto   - Autoindent state          (flip)
    //   AC    - Autocompletion            (flip)
    //  Mark   - Marking state             1 | 0
    //  Row    - Current absolute Row
    //  Col    - Current absolute Col
    //  Page   - Codepage
    //  Buffer - switch to buffer with handle
    //  Name   - Current buffer name
    //  WW     - WordWrap Width
    //  WS     - WordWrap Status
    //  WM     - WordWrap Merge
    //  WP     - WordWrap Para
    //  T      - tab width

    switch(__to_upper(argv[0].strptr[0]))
    {
        case 'T':
            	if(lCnt != 2)
            	    return makeRC(RC, -3);

            	{
            		ULONG ulData = s2u((char*)argv[1].strptr);

            		if(ulData > 0 && ulData <= 8)
            		{
						iTabWidth = ulData;
						Editor.recalc_sz();
					}
            		else
            		    return makeRC(RC, -3);
            	}

        		return makeRC(RC, 0);

        case 'I':
                if(lCnt == 2 && argv[1].strlength == 1)
                {
                    switch(argv[1].strptr[0])
                    {
                        case '0':
                        case '1':
                            Editor.current()->set_ins_mode(argv[1].strptr[0] - '0');
                            return makeRC(RC, 0);

                        default:
                            return makeRC(RC, -3);
                    }
                }
                if(lCnt == 2)
                    return makeRC(RC, -3);

                Editor.current()->set_ins_mode(1 - Editor.current()->get_ins_mode());

                return makeRC(RC, 0);

        case 'A':

                if(argv[0].strlength > 1 && __to_upper(argv[0].strptr[1] == 'C'))
                {
                    if(lCnt == 2 && argv[1].strlength == 1)
                    {
                        switch(argv[1].strptr[0])
                        {
                            case '0':
                            case '1':
                                Editor.current()->set_auto_completion(argv[1].strptr[0] - '0');
                                return makeRC(RC, 0);

                            default:
                                return makeRC(RC, -4);
                        }
                    }
                    if(lCnt == 2)
                        return makeRC(RC, -4);

                    Editor.current()->set_auto_completion(1 - Editor.current()->get_auto_completion());

                    return makeRC(RC, 0);
                }

                if(lCnt == 2 && argv[1].strlength == 1)
                {
                    switch(argv[1].strptr[0])
                    {
                        case '0':
                        case '1':
                            Editor.current()->set_auto_indent(argv[1].strptr[0] - '0');
                            return makeRC(RC, 0);

                        default:
                            return makeRC(RC, -4);
                    }
                }
                if(lCnt == 2)
                    return makeRC(RC, -4);

                Editor.current()->set_auto_indent(1 - Editor.current()->get_auto_indent());

                return makeRC(RC, 0);

        case 'H':

                if(lCnt == 2 && argv[1].strlength == 1)
                {
                    switch(argv[1].strptr[0])
                    {
                        case '0': case '1': case '2': case '3': case '4':
                        case '5': case '6': case '7': case '8': case '9':
                            Editor.current()->set_hiliting(argv[1].strptr[0] - '0');
                            return makeRC(RC, 0);

                        default:
                            return makeRC(RC, -5);
                    }
                }
                if(lCnt == 2)
                    return makeRC(RC, -5);

                Editor.current()->flip_hiliting();

                return makeRC(RC, 0);


        //For listed below options, additional parameter is required
        case 'M':

                if(lCnt == 1)
                {
                    Editor.current()->unmark();
                    return makeRC(RC, 0);
                }

                if(argv[1].strlength == 1)
                {
                    if(argv[1].strptr[0] == '1')
                    {
                        Editor.current()->mark();
                        return makeRC(RC, 0);
                    }
                    else
                    {
                        Editor.current()->unmark();
                        return makeRC(RC, 0);
                    }
                }
                return makeRC(RC, -6);

        case 'R':
                if(lCnt == 2 && argv[1].strlength > 0)
                {
                    ULONG ulData = s2u((char*)argv[1].strptr);

                    if(ulData)
                    {
                        Editor.current()->goto_line(ulData - 1);
                        return makeRC(RC, 0);
                    }
                }

                return makeRC(RC, -7);

        case 'C':

                if(lCnt == 2 && argv[1].strlength > 0)
                {
                    ULONG ulData = s2u((char*)argv[1].strptr);

                    if(ulData)
                    {
                        Editor.current()->goto_col(ulData - 1);
                        return makeRC(RC, 0);
                    }
                }

                return makeRC(RC, -8);

        case 'B':
                if(lCnt == 2 && argv[1].strlength > 0)
                {
                    EditBox* new_box = (EditBox*)s2u((char*)argv[1].strptr);
                    if(new_box)
                    {
                        Editor.select(new_box);
                        return makeRC(RC, 0);
                    }
                }

                return makeRC(RC, -9);

        case 'N':
                if(lCnt == 2 && argv[1].strlength > 0)
                    Editor.current()->set_name((char*)argv[1].strptr);
                else
                    return makeRC(RC, -10);

                return makeRC(RC, 0);

        case 'P':
                if(lCnt == 2 && argv[1].strlength > 0)
                    Editor.current()->set_xlate((char*)argv[1].strptr);
                else
                    return makeRC(RC, -11);

                return makeRC(RC, 0);

        case 'W':
                {
                    if(argv[0].strlength != 2 || lCnt != 2)
    	               break;

                    Buffer* p = Editor.current();
                    int flags = p->ww_get();
                    int fnew  = 0;
                    int val   = s2u((char*)argv[1].strptr);

                    switch(__to_upper(argv[0].strptr[1]))
                    {
                	    case 'W':
                            p->ww_set_width(val);
                            return makeRC(RC, 0);

                        case 'S': fnew = WW_STATE; break;
                        case 'M': fnew = WW_MERGE; break;
                        case 'P': fnew = WW_LONG ; break;
                    }

                    if(val)
                        flags |= fnew;
                    else
                        flags &= ~fnew;

                    p->ww_set(flags);

                    return makeRC(RC, 0);
                }
        	    break;
    }

    return makeRC(RC, -2);
}

ULONG APIENTRY fedFind(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    if(lCnt < 2)
        return makeRC(RC, -1);

    char* repl = (char*)((lCnt > 2 && argv[2].strptr) ? (char*)argv[2].strptr:"");

    int match_len = Editor.current()->search((char*)argv[1].strptr,
                                             (char*)argv[0].strptr,
                                             repl);

    return makeRC(RC, match_len);
}

ULONG APIENTRY fedMenu(PCSZ, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    int r = 0;
    int c = 0;

    if(lCnt < 3 ||
       !argv[0].strlength ||
       !argv[1].strlength ||
       !argv[2].strlength)
    {
        return makeRC(RC, -1);
    }

    r = s2u((char*)argv[0].strptr);
    c = s2u((char*)argv[1].strptr);

    int choice = -1;
    int i;

    Dialog dlg(r, c, 1, 1);
    Menu* menu = new Menu(1, 1, 0);

    for(i = 2; i < lCnt; i++)
    {
        if(argv[i].strlength)
            menu->add_menu((char*)argv[i].strptr);
    }

    dlg.Ins(menu);
    dlg.rows = menu->rows+2;
    dlg.cols = menu->cols+2;

    KeyInfo k;
    do
    {
        dlg.draw();
        vio_read_key(&k);

        if((k.skey & 0x00FF) == kbEnter || (k.skey & 0x00FF) == kbGrEnter)
        {
            choice = menu->first_selected();
            break;
        }

        dlg.do_key(k);

        if(menu->selected_by_key())
        {
            choice = menu->first_selected();
            break;
        }
    }
    while(k.skey != (kbEsc | shIsCtrl));

    return makeRC(RC, choice + 1);
}

ULONG APIENTRY fedJumpList(PCSZ pName, ULONG lCnt, PRXSTRING argv, PCSZ, PRXSTRING RC)
{
    char varname[512];
    char varvalue[512];
    char header[512];
    APIRET rc;
    int len;
    int i;
    int list;

    if(lCnt < 1 || argv[0].strlength == 0 || !argv[0].strptr)
        return makeRC(RC, -1);

    if(!strcmp((char*)pName, "FEDOPENJUMPLIST"))
    {
        if(lCnt < 1)
            return makeRC(RC, -2);

        list = s2u((char*)argv[0].strptr);

        if(list > 9)
            return makeRC(RC, -3);

        int r  = 0;
        int c  = 0;
        int nr = 0;
        int nc = 0;

        if(lCnt >= 2)
            r = s2u((char*)argv[1].strptr);
        if(lCnt >= 3)
            c = s2u((char*)argv[2].strptr);
        if(lCnt >= 4)
            nr = s2u((char*)argv[3].strptr);
        if(lCnt >= 5)
            nc = s2u((char*)argv[4].strptr);

        return makeRC(RC, JumpListBox(&JList[list], r, c, nr, nc));
    }

    if(lCnt != 2)
        return makeRC(RC, -2);

    list = s2u((char*)argv[1].strptr);

    if(list > 9)
        return makeRC(RC, -3);

    rc = GetRexxVariable(mk_var(varname, (char*)argv[0].strptr,0,"jump_len",0),
                         varvalue, sizeof(varvalue)-1);

    if(rc)
        return makeRC(RC, -4);

    len = s2u(varvalue);

    JList[list].RemoveAll();

    if(!len)
        return makeRC(RC, 0);

    int new_pos = -1;

    rc = GetRexxVariable(mk_var(varname, (char*)argv[0].strptr,0,"jump_start",0),
                         varvalue, sizeof(varvalue)-1);

    if(!rc)
        new_pos = s2u(varvalue);

    for(i = 0; i < len; i++)
    {
        int row;
        int col;
        APIRET rc2 = 1;

        rc = GetRexxVariable(mk_var(varname,(char*)argv[0].strptr, i, "jump_row", 1),
                             varvalue, sizeof(varvalue)-1);

        if(rc)
            return makeRC(RC, -5);

        row = s2u(varvalue);

        rc = GetRexxVariable(mk_var(varname,(char*)argv[0].strptr, i, "jump_col", 1),
                             varvalue, sizeof(varvalue)-1);

        if(rc)
            return makeRC(RC, -6);

        col = s2u(varvalue);

        rc2 = GetRexxVariable(mk_var(varname,(char*)argv[0].strptr, i, "jump_header", 1),
                             header, sizeof(header)-1);

        rc = GetRexxVariable(mk_var(varname,(char*)argv[0].strptr, i, "jump_file", 1),
                             varvalue, sizeof(varvalue)-1);
        if(rc)
            return makeRC(RC, -7);

        if(rc2)
            strcpy(header, varvalue);

        JList[list].add_entry(row, col, header, varvalue);
    }

    JList[list].set_last_pos(new_pos);

    return makeRC(RC, 0);
}



//-----------------------------------------
// Utility functions
//-----------------------------------------

int makeRC(PRXSTRING RC, LONG lData)
{
	RC->strlength = strlen(i2s(lData, 0, (char*)RC->strptr));
    return 0;
}

int makeRCstr(PRXSTRING RC, char *str)
{
    if(!str)
        return 0;

    makeRCstr(RC, strlen(str));
    strcpy((char*)RC->strptr, str);
    RC->strlength = strlen(str);
    return 0;
}

int makeRCstr(PRXSTRING RC, int Len)
{
    if(Len > RC->strlength)
    {
        DosFreeMem(RC->strptr);

        DosAllocMem((PPVOID)&RC->strptr, Len, PAG_COMMIT | PAG_WRITE);
    }

    RC->strlength = Len;
    return 0;
}

APIRET SetRexxVariable(char* name, char* value, LONG len)
{
	if(!name || !value)
		return 1;

	if(len < 0)
		len = strlen(value);

    SHVBLOCK block = {0};

	block.shvcode = RXSHV_SYSET;

    MAKERXSTRING(block.shvname, name, strlen(name));
    MAKERXSTRING(block.shvvalue, value, len);

    block.shvnamelen  = block.shvname.strlength;
	block.shvvaluelen = block.shvvalue.strlength;

	return RexxVariablePool(&block);
}

APIRET GetRexxVariable(char* name, char* value, LONG len)
{
    APIRET rc;
    SHVBLOCK block = {0};

    block.shvcode = RXSHV_SYFET;

    MAKERXSTRING(block.shvname, name, strlen(name));
    MAKERXSTRING(block.shvvalue, value, len);

    block.shvvaluelen=len;

    rc = RexxVariablePool(&block);

    if(!rc)
        value[block.shvvalue.strlength]=0;
    else
        value[0]=0;
    return rc;
}

char* mk_var(char *buff, char *beg, int num, char *end, int mode)
{
    strcpy(buff, beg);

    if(mode)
    {
    	char cNum[16];

        strcat(buff, ".");
        strcat(buff, u2s(num,0, cNum));
    }
    strcat(buff, ".");
    strcat(buff, end);
    return buff;
}

